#include "SkillImplementationWrapper.h"

namespace armarx
{
    namespace skills::detail
    {
        SkillRuntime::SkillRuntime(
            const skills::SkillBlueprint& fac,
            const skills::SkillExecutionID& execId,
            const aron::data::DictPtr& initial_parameters,
            const skills::callback::dti::SkillProviderCallbackInterfacePrx& callbackInterface) :
            factory(fac), statusUpdate{{execId, initial_parameters, callbackInterface}}
        {
        }

        // ask a skill to stop
        void
        SkillRuntime::stopSkill()
        {
            while (this->skill == nullptr)
            {
                ARMARX_INFO << deactivateSpam()
                            << "Trying to stop a skill that has not been constructed yet... "
                               "Waiting until construction...";
                std::this_thread::sleep_for(std::chrono::milliseconds(100));
            }

            this->skill->notifySkillToStop();
        }

        void
        SkillRuntime::updateSkillParameters(const aron::data::DictPtr& i)
        {
            while (this->skill == nullptr)
            {
                ARMARX_INFO << deactivateSpam()
                            << "Trying to add params to a skill that has not been constructed "
                               "yet... Waiting until construction...";
                std::this_thread::sleep_for(std::chrono::milliseconds(100));
            }

            this->skill->updateParameters(i);
        }

        TerminatedSkillStatusUpdate
        SkillRuntime::executeSkill()
        {
            std::scoped_lock l(this->executionMutex);
            // -------------------------------------------------------------------------------------
            // sanity check
            // -------------------------------------------------------------------------------------
            ARMARX_CHECK(this->skill == nullptr)
                << "A SKILL HAS ALREADY BEEN CONSTRUCTED. THIS SHOULD NOT HAPPEN!!";


            // -------------------------------------------------------------------------------------
            // setup basic vars and lambdas
            // -------------------------------------------------------------------------------------
            // actually we should lock... however this is only read access and the members are const throughout the execution...
            ARMARX_CHECK(statusUpdate.executionId.skillId.isFullySpecified());

            const auto& initial_aron_params = statusUpdate.parameters;
            const auto& callback_interface = statusUpdate.callbackInterface;
            const auto& skillId = statusUpdate.executionId.skillId;
            const auto& skillName = skillId.skillName;
            const auto& providerId = *skillId.providerId;
            const auto& executorName = statusUpdate.executionId.executorName;
            const auto& manager =
                skills::manager::dti::SkillManagerInterfacePrx::checkedCast(callback_interface);

            ARMARX_INFO_S << "Executing skill: " << skillName;

            auto updateStatus =
                [&](const SkillStatus status, const aron::data::DictPtr& data = nullptr)
            {
                std::unique_lock l(skillStatusesMutex);
                statusUpdate.status = status;
                statusUpdate.result = data;
                if (skill) // if skill has been constructed
                {
                    // update parameterization
                    statusUpdate.parameters = skill->getParameters();
                }

                if (callback_interface) // if callback interface is used
                {
                    auto pid = providerId.toCallbackIce();
                    callback_interface->updateStatusForSkill(statusUpdate.toProviderIce(), pid);
                }
            };

            auto createErrorMessageData = [](const std::string& message)
            {
                auto obj = aron::make_dict();
                auto m = aron::make_string(message, aron::Path({"errormessage"}));
                obj->addElement("errormessage", m);
                return obj;
            };

            {
                std::unique_lock l(skillStatusesMutex);

                // ---------------------------------------------------------------------------------
                // check times
                // ---------------------------------------------------------------------------------
                if ((armarx::core::time::DateTime::Now() -
                     statusUpdate.executionId.executionStartedTime) >
                    armarx::core::time::Duration::Seconds(1))
                {
                    ARMARX_WARNING_S << "SkillError 001: For some reason the skill '"
                                     << skillId.toString()
                                     << "' has been scheduled > 1s ago. The execution should start "
                                        "much faster! Continue, but this is bad behavior..";
                }
            }

            // -------------------------------------------------------------------------------------
            // construct skill
            // -------------------------------------------------------------------------------------
            ARMARX_INFO_S << "Construct skill: " << skillName;

            updateStatus(SkillStatus::Constructing);
            this->skill = this->factory.createSkill(providerId);
            this->skill->setExecutorName(executorName);
            this->skill->setManager(manager);
            this->skill->setCallback([&](const SkillStatus s, const armarx::aron::data::DictPtr& d)
                                     { updateStatus(s, d); });

            // set initial parameters that were attached to the execution request (only add as we are not sure whether some updates already arrived)
            this->updateSkillParameters(initial_aron_params);

            auto makeAbortedResult = [&](const std::string& message)
            {
                updateStatus(SkillStatus::Aborted, createErrorMessageData(message));

                std::unique_lock l(skillStatusesMutex);
                auto terminated = TerminatedSkillStatusUpdate{
                    {.executionId = statusUpdate.executionId,
                     .parameters = statusUpdate.parameters,
                     .callbackInterface = statusUpdate.callbackInterface}};
                terminated.status = TerminatedSkillStatus::Aborted;
                return terminated;
            };


            auto makeFailedResult = [&](const std::string& message)
            {
                updateStatus(SkillStatus::Failed, createErrorMessageData(message));

                std::unique_lock l(skillStatusesMutex);
                auto terminated = TerminatedSkillStatusUpdate{
                    {.executionId = statusUpdate.executionId,
                     .parameters = statusUpdate.parameters,
                     .callbackInterface = statusUpdate.callbackInterface}};
                terminated.status = TerminatedSkillStatus::Failed;
                return terminated;
            };

            // -------------------------------------------------------------------------------------
            // Check params
            // -------------------------------------------------------------------------------------
            // NOT RELEVANT ANYMORE!!
            //            if (skill->getSkillDescription().acceptedType && initial_aron_params &&
            //                not(initial_aron_params->fullfillsType(skill->getSkillDescription().acceptedType)))
            //            {
            //                std::string message =
            //                    "SkillError 002: The Skill '" + skillName +
            //                    "' has a type and got parameters but the input does not match the type.";
            //                ARMARX_ERROR_S << message;
            //                return makeTerminationResult(message);
            //            }


            auto exitAndMakeFailedResult = [&](const std::string& message)
            {
                skill->exitSkill(); // try to exit skill. Ignore return value
                return makeFailedResult(message);
            };
            auto exitAndMakeAbortedResult = [&](const std::string& message)
            {
                skill->exitSkill(); // try to exit skill. Ignore return value
                return makeAbortedResult(message);
            };

            // Construction succeeded!
            // -------------------------------------------------------------------------------------
            // Init skill
            // -------------------------------------------------------------------------------------
            ARMARX_INFO_S << "Init skill: " << skillName;
            updateStatus(SkillStatus::Initializing);

            try
            {
                Skill::InitResult initRet = skill->initSkill();
                if (initRet.status != TerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 101: The initialization of skill '" +
                                          skillName + "' did not succeed.";
                    return exitAndMakeFailedResult(message);
                }
            }
            catch (const error::SkillAbortedException& ex)
            {
                return exitAndMakeAbortedResult(GetHandledExceptionString());
            }
            catch (const error::SkillFailedException& ex)
            {
                return exitAndMakeFailedResult(GetHandledExceptionString());
            }
            catch (const std::exception& ex)
            {
                std::string message =
                    "SkillError 101e: An error occured during the initialization of skill '" +
                    skillName + "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return exitAndMakeFailedResult(message);
            }

            // Init succeeded!
            // -------------------------------------------------------------------------------------
            // Prepare skill
            // -------------------------------------------------------------------------------------
            ARMARX_INFO_S << "Prepare skill: " << skillName;
            updateStatus(SkillStatus::Preparing);

            try
            {
                auto prepareRet = skill->prepareSkill();
                while (prepareRet.status == ActiveOrTerminatedSkillStatus::Running)
                {
                    ARMARX_INFO << deactivateSpam() << "Not all requirements for skill "
                                << skillName << " fulfilled. Waiting...";

                    // wait...
                    std::this_thread::sleep_for(std::chrono::milliseconds(50));
                    prepareRet = skill->prepareSkill();
                }

                if (prepareRet.status != ActiveOrTerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 201: The prepare method of skill '" +
                                          skillName + "' did not succeed.";
                    return exitAndMakeFailedResult(message);
                }
            }
            catch (const error::SkillAbortedException& ex)
            {
                return exitAndMakeAbortedResult(GetHandledExceptionString());
            }
            catch (const error::SkillFailedException& ex)
            {
                return exitAndMakeFailedResult(GetHandledExceptionString());
            }
            catch (const std::exception& ex)
            {
                std::string message = "SkillError 201e: An error occured during waiting for skill "
                                      "dependencies of skill '" +
                                      skillName +
                                      "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return exitAndMakeFailedResult(message);
            }

            // Prepare succeeded!
            // -------------------------------------------------------------------------------------
            // Main of skill
            // -------------------------------------------------------------------------------------
            // execute. If the skill fails for some reason, from this point it will always execute its exit function.
            ARMARX_INFO_S << "Main skill: " << skillName;
            updateStatus(SkillStatus::Running);

            Skill::MainResult mainRet;
            try
            {
                mainRet = skill->mainOfSkill();
                if (mainRet.status != TerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 501: The main method of skill '" + skillName +
                                          "' did not succeed.";
                    return exitAndMakeFailedResult(message);
                }
            }
            catch (const error::SkillAbortedException& ex)
            {
                return exitAndMakeAbortedResult(GetHandledExceptionString());
            }
            catch (const error::SkillFailedException& ex)
            {
                return exitAndMakeFailedResult(GetHandledExceptionString());
            }
            catch (const std::exception& ex)
            {
                std::string message =
                    "SkillError 501e: An error occured during the main method of skill '" +
                    skillName + "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return exitAndMakeFailedResult(message);
            }

            // Main succeeded!
            // -------------------------------------------------------------------------------------
            // Exit of skill
            // -------------------------------------------------------------------------------------
            ARMARX_INFO_S << "Exit skill: " << skillName;
            try
            {
                Skill::ExitResult exitRet = skill->exitSkill();
                if (exitRet.status != TerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 601: The exit method of skill '" + skillName +
                                          "' did not succeed.";
                    return makeFailedResult(message);
                }
            }
            catch (const error::SkillAbortedException& ex)
            {
                return makeAbortedResult(GetHandledExceptionString());
            }
            catch (const error::SkillFailedException& ex)
            {
                return makeFailedResult(GetHandledExceptionString());
            }
            catch (const std::exception& ex)
            {
                std::string message =
                    "SkillError 601e: An error occured during the exit method of skill '" +
                    skillName + "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return makeFailedResult(message);
            }

            // Exit succeeded!
            // All succeeded!
            {
                updateStatus(SkillStatus::Succeeded);

                // return result of main method
                std::unique_lock l(skillStatusesMutex);
                TerminatedSkillStatusUpdate ret{
                    {.executionId = statusUpdate.executionId,
                     .parameters = statusUpdate.parameters,
                     .callbackInterface = statusUpdate.callbackInterface}};
                ret.result = mainRet.data;
                ret.status = TerminatedSkillStatus::Succeeded;
                return ret;
            }
        }
    } // namespace skills::detail
} // namespace armarx
